package provide mmuclGui 0.5
package require form
package require balloon
package require htmlHelp

#########################################################################
# mmuclGui.tcl - graphical interfaces to core mmucl procedures
#
# Copyright (C) 1997-1998 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#########################################################################

global mmucl
array set mmucl {
    echo            1
    echo_attribs    {yellow}
    keep_line       0
    browser         internal
    hist_min_length 3
}

namespace eval mmucl {
    namespace export guiInit userEntry optionMenu toolbar
}

# mmucl::guiInit --
#
#   Load mmuclGui.conf, images...

proc mmucl::guiInit {} {
    variable Priv
    global env tcl_platform

    array set Priv {
	histloc 1
    }
    set Priv(lib_dir) $env(MMUCL_DIR)   

    # The image name will be the same as the file name
    foreach image [glob [file join $Priv(lib_dir) icons *.gif]] {
	image create photo [file tail $image] -file $image
    }

    # Specify an icon. Won't work on some platforms.
    if {$tcl_platform(platform) == "unix"} {
	toplevel .icon
	label .icon.l -image mmucl.gif -bd 0
	pack .icon.l
	wm iconwindow . .icon
    }
    
    # Load gui configuration.
    set conf [file join $Priv(lib_dir) mmuclGui.conf]
    if {[file exists $conf]} {
	if {[catch {option readfile $conf} error]} {
	    puts stderr "Error in mmuclGui.conf: $error"
	}
    }
}

# mmucl::userEntry --
#
#   Set up an entry widget to parse input to mmucl.

proc mmucl::userEntry {w} {
    entry $w
    
    bind $w <Key-Return> {mmucl::meta_parse %W}
    bind $w <Key-Up> {mmucl::histScroll %W -1}
    bind $w <Key-Down> {mmucl::histScroll %W 1}
    
    # Toggle a debugging shell
    bind $w <Control-Shift-Escape> {
	if {[string match *meta_parse* [bind %W <Key-Return>]]} {
	    mmucl::MCecho "Entering debug mode" bold
	    bind %W <Key-Return> {mmucl::debug_parse %W}
	} else {
	    mmucl::MCecho "Leaving debug mode" bold
	    bind %W <Key-Return> {mmucl::meta_parse %W}
	}
    }

    bindtags $w [list $w mmucl Entry . all]
    return $w 
}

# mmucl::toolbar --
#
#   Make a toolbar.

proc mmucl::toolbar {w} {
    frame $w
    
    button $w.con -command mmucl::tkConnect -image connect.gif 
    balloon_add $w.con "Connect to mud"
    button $w.load -image load.gif  \
	-command {tk_openFile mmucl::msource -title {Load a File}}
    balloon_add $w.load "Load a file"
    button $w.send -image send.gif \
	-command {tk_openFile mmucl::MCtextin -title {Send a file to the mud}}
    balloon_add $w.send "Send a file"
    button $w.save -image save.gif \
	-command {tk_saveFile save -title {Save Current State}}
    balloon_add $w.save "Save state"
    button $w.char -image char.gif -command {mmucl::tkEdit Char}
    balloon_add $w.char "Edit characters"
    button $w.alias -image alias.gif -command {mmucl::tkEdit Alias}
    balloon_add $w.alias "Edit aliases"
    button $w.action -image action.gif -command {mmucl::tkEdit Action }
    balloon_add $w.action "Edit actions"
    button $w.bind -image bind.gif -command {mmucl::tkEdit Bind}
    balloon_add $w.bind "Edit key bindings"
    button $w.sub -image sub.gif -command {mmucl::tkEdit Sub}
    balloon_add $w.sub "Edit substitutions"

    pack $w.con $w.load $w.send $w.save $w.char $w.alias $w.action $w.bind\
	$w.sub -side left 
    return $w
}

# mmucl::optionMenu --
#
#   Make a menu of mmucl options and return it.

proc mmucl::optionMenu {w} {
    
    menu $w
    $w add checkbutton -label Reconnect -variable mmucl(reconnect)
    $w add checkbutton -label "Keep line" \
	-variable mmucl(keep_line)
    $w  add checkbutton -label "Local echo" -variable mmucl(echo)
    $w add checkbutton -label "No actions" \
	-variable mmucl(no_action)
    $w add checkbutton -label "No subs" -variable mmucl(no_sub)
    $w  add cascade -label "Echo color" \
	-menu $w.colors
    
    menu $w.colors
    foreach color {black red green yellow blue magenta cyan grey} {
	$w.colors add radiobutton -label $color -value $color \
	    -variable mmucl(echo_attribs)
    }
    return $w
}

# mmucl::meta_parse --
#
#   A layer on top of parse that adds echoing, echo coloring, optionally
#   records input in history. Should be bound to an entry widget.

proc mmucl::meta_parse {w} {
    global mmucl
    variable Priv

    set input [$w get]

    if {[string compare $mmucl(echo) 0]} {
	MCecho $input [split $mmucl(echo_attribs)]
    }

    MCparse $input

    if {[string length $input] >= $mmucl(hist_min_length)} {
	history add $input
	set Priv(histloc) [history nextid]
    }

    if {[string compare $mmucl(keep_line) 0]} {
	$w selection range 0 end
    } else {
	$w delete 0 end
    }
}

# mmucl::debug_parse --
#
#   Used to make a debugging shell.

proc mmucl::debug_parse {w} {
    variable Priv

    set input [$w get]
    
    MCecho $input {bold green}
    set color {grey bold}
    if {[catch {set result [uplevel \#0 $input]} result]} {
	set color {bold red}
    }
    MCecho $result $color
    history add $input
    set Priv(histloc) [history nextid]
    $w delete 0 end
}

# mmucl::histScroll --
#
#   Scroll through history in an entry widget.
#   Save and restore the working line as needed.

proc mmucl::histScroll {w dir} {
    variable Priv
 
    set keep [history keep]
    set nid [history nextid]

    if {$Priv(histloc) == $nid} {
	set Priv(wline) [$w get]
    }

    catch {history change [$w get] $Priv(histloc)}

    incr Priv(histloc) $dir
    if {$Priv(histloc)>$nid || $Priv(histloc)<$nid-$keep || $Priv(histloc)<1} {
	incr Priv(histloc) -$dir
    } else {
	if {$Priv(histloc) == $nid} {
	    set text $Priv(wline)
	} else {
	    set text [history event $Priv(histloc)]
	}
	$w delete 0 end
	$w insert end $text
    }
}

# mmucl::MCkb_bind --
#
#   Let the user bind commands to .mmucl_input. The command are executed in
#   the mmucl interp.

proc mmucl::MCkb_bind {{event *} {script ""}} {
    variable Bind

    if {$script == ""} {
	set result [array names Bind $event]
	if {$result != ""} {
	    foreach match [lsort $result] {
		MCecho "** Event $match: $Bind($match) **"
	    }
	} else {
	    MCecho "** No binding matching: $event **"
	}
    } else {
	if {[catch {bind mmucl $event \
			[list mmucl::meval $script "event $event"]} error]} {
	    MCecho "Error: $error" red
	} else { 
	    set Bind($event) [list $script]
	    MCecho "** Event $event bound to {$script} **"
	}
    }
}

# mmucl::MCkb_unbind --
#
#   Delete a user defined binding

proc mmucl::MCunkb_bind {pattern} {
    variable Bind
    
     # disable matching of []
    regsub -all {\[} $pattern {\\&} pattern

    set result [array names Bind $pattern]
    if {$result !=""} {
	foreach match $result {
	    unset Bind($match)
	    bind mmucl $match {}
	    MCecho "** Binding to event $match deleted. **"
	}
    } else {
	MCecho "** No binding matching $pattern. **"
    }
}

# mmucl::MCloadTk --
#
#   Load Tk into interp mmucl making sure the right
#   globals vars are set. It also handles safe interpreters.
#   Note that on safe interpreters loadTk makes the toplevel
#   and wm commands available on safe interps.

proc mmucl::MCloadTk {} {
    global tk_library env

    mmucl eval [list set tk_library $tk_library]
    
    if {[interp issafe mmucl]} {
	interp expose mmucl source
	interp alias mmucl file {} file
	load {} Tk mmucl
	interp hide mmucl source
	interp alias mmucl file {} {}
	interp expose mmucl toplevel
	interp expose mmucl wm
    } else {
	load {} Tk mmucl
    }
    mmucl eval {wm withdraw .}
}


# mmucl::MChelp --
#
#   Spawns a user defined html browser.

proc mmucl::MChelp {{subject contents}} {
    global mmucl
    variable Priv
    
    set file [file join $Priv(lib_dir) html $subject.html]
    if {![file exists $file]} {
 	# should go to some kind of error page
 	set file [file join $Priv(lib_dir) html contents.html]
    }
    # Fix up windows paths.
    regsub -all {\\} $mmucl(browser) / browser

    if {$browser=="internal"} {
	tk_htmlHelp $file
    } else {
	exec $browser $file &
    }
}

# mmucl::tkConnect --
#
#   Construct a dialog box for connecting to a mud either by
#   selecting a defined character or typing the location and port in.
#   Ugly.

proc mmucl::tkConnect {} {
    variable Char
    variable Priv
    
    if {[winfo exists .con]} {raise .con; return}
    
    toplevel .con
    wm title .con "Connect"

    frame .con.f -relief groove -bd 2
    grid [label .con.f.hostl -text "Host:"]
    grid [entry .con.f.host] -row 0 -column 1
    grid [label .con.f.portl -text "Port:"] 
    grid [entry .con.f.port] -row 1 -column 1
    
    .con.f.host insert end $Priv(host)
    .con.f.port insert end $Priv(port)
    set Priv(tmpchar) ""

    # if characters defined, make a character selection area
    if {[array size Char]} {
	grid [label .con.f.lchar -text "Character:"]
	set chars [lsort [array names Char]]
        set m [eval tk_optionMenu .con.f.char mmucl::Priv(tmpchar) $chars]
	grid .con.f.char -row 2 -column 1
	$m add separator
	$m add command -label None -command {set mmucl::Priv(tmpchar) ""}
	set Priv(tmpchar) ""
    }
    
    frame .con.b
    pack [button .con.b.ok -text Connect \
	      -command {
		  if {$mmucl::Priv(tmpchar) == ""} {
		      mmucl::MCconnect [.con.f.host get] [.con.f.port get]
		  } else {mmucl::MCcharload $mmucl::Priv(tmpchar)}
		  destroy .con}
	 ] -side left
    
    pack [button .con.b.quit -text Cancel -command {destroy .con}] -side left
  
    pack .con.f .con.b
    wm geometry .con +[winfo pointerx .]+[winfo pointery .]
    wm resizable .con 0 0
}

# mmucl::tkEdit --
#
#   Pops up a dialog box to let the user edit the given array.  

proc mmucl::tkEdit {arrayName} {
    switch $arrayName {
	Action {
	    set w .action 
	    set labels {Pattern: Script:}
	    set add action
	    set delete unaction
	    set title "Edit Actions"
	} Alias {
	    set w .alias 
	    set labels {Name: Script:}
	    set add alias
	    set delete unalias
	    set title "Edit Aliases"
	} Bind {
	    set w .bind
	    set labels {Event: Script:}
	    set add kb_bind
	    set delete unkb_bind
	    set title "Edit Key Bindings"
	} Char {
	    set w .char
	    set labels {Name: Mud: Host: Port: Login:}
	    set add charadd
	    set delete chardelete
	    set title "Edit Characters"
	} Sub {
	    set w .sub
	    set labels {Pattern: Replacement:}
	    set add sub
	    set delete unsub
	    set title "Edit Substitutions"
	}
    }

    if {[winfo exists $w]} {
	raise $w
	return
    }
  
    toplevel $w
    wm title $w $title

    form $w.form -array mmucl::$arrayName -label $labels -add mmucl::MC$add \
	-del mmucl::MC$delete
    pack $w.form -side top -pady 3
 
    frame $w.b
    pack [button $w.b.help -text Help -command [list mmucl::MChelp $add]] \
  	[button $w.b.exit -text Exit -command [list destroy $w]] \
	-side left
    pack $w.b -side bottom
    
    wm geometry $w +[winfo pointerx .]+[winfo pointery .]
    wm resizable $w 0 0
}


# Right click in any entry widget to insert the selection or else 
# the clipboard if present.
bind Entry <Button-3> {
    if {[catch {selection get}]} {
	catch {%W insert insert [selection get -selection CLIPBOARD]}
    } else {
	%W insert insert [selection get]
    }
}


# Misc procedures.

# Dialog for giving an existing file as an arg to a command

proc tk_openFile {proc args} {
    set filename  [eval tk_getOpenFile $args]
    
     if {$filename != ""} {
	 $proc $filename
     }
}

# Dialog for giving a file to a command as an arg

proc tk_saveFile {proc args} {
    set filename [eval tk_getSaveFile $args]
   
    if {$filename != ""} {
	$proc $filename
    }
}

# Execute a proc when given some result from tk_messageBox

proc tk_dialogBox {proc on_result args} {
    set result [eval tk_messageBox $args]

    if {$result == $on_result} {
	$proc
    }
}
