package provide textHtml 0.12

##########################################################################
# textHtml.tcl - A package based on html_library 0.3 by Stephel Uhler for
#         displaying html in a text widget.
#
# Copyright (C) 1997-1998 Mark Patton
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
##########################################################################

# This is deliberately a very minimalist package for displaying
# simple html. No tables, forms, images etc supported. 

namespace eval textHtml {
    namespace export init_win reset_win

    variable tag_map
    variable insert_map
    variable list_elements

    # mapping of html tags to text tag properties
    # properties beginning with "T" map directly to text tags
    
    array set tag_map {
	b      {weight bold}
	blockquote   {indent 1 Trindent rindent}
	bq     {style italic indent 1 Trindent rindent}
	cite   {style italic}
	code   {family courier}
	dfn    {style italic}	
	dir    {indent 1}
	dd     {indent 1}
	em     {style italic}
	h1     {size 18 weight bold}
	h2     {size 16 weight bold}		
	h3     {size 14 weight bold}	
	h4     {size 14}
	h5     {size 12}
	h6     {style italic}
	i      {style italic}
	kbd    {family courier weight bold}
	menu   {indent 1}
	ol     {indent 1}
	pre    {fill 0 family courier Tnowrap nowrap}
	samp   {family courier}		
	strong {weight bold}
	tt     {family courier}
	u      {Tunderline underline}
	ul     {indent 1}
	var    {style italic}
	address {style italic}
}

    # These are in common(?) use, but not defined in html2.0
    
    array set tag_map {
	center {Tcenter center}
	strike {Tstrike strike}
	u      {Tunderline underline}
    }

    # initial values
    
    set tag_map(hmstart) {
	family     Times
	size       12
	weight     ""
	style      ""
	Tcenter    ""
	Tlink      ""
	Tnowrap    ""
	Tunderline ""
	list       list
	fill       1
	indent     ""
	counter    0
	adjust     0
    }
    
    # html tags that insert white space
    
    array set insert_map {
        br      "\n"
        dd      "\n"      /dd         "\n"
        dl      "\n"      /dl         "\n"
        dt      "\n"
        form    "\n"      /form       "\n"
        h1      "\n\n"    /h1         "\n"
        h2      "\n\n"    /h2         "\n"
        h3      "\n\n"    /h3         "\n"
        h4      "\n\n"    /h4         "\n"
        h5      "\n"      /h5         "\n"
        h6      "\n"      /h6         "\n"
        li      "\n"
        /dir    "\n"
        /ul     "\n"
        /ol     "\n"
        /menu   "\n"
        p       "\n\n"
        pre     "\n"      /pre        "\n"
        /address    "\n"
    }
    
    # tags that are list elements, that support "compact" rendering
    
    array set list_elements {
	ol    1
	ul    1
	menu  1
	dl    1
	dir   1
    }
}


############################################
# initialize the window and stack state

proc textHtml::init_win {win} {
    upvar #0 textHtml::$win var

    init_state $win
    $win tag configure underline -underline 1
    $win tag configure center -justify center
    $win tag configure nowrap -wrap none
    $win tag configure rindent -rmargin $var(S_tab)c
    $win tag configure strike -overstrike 1
    $win tag configure mark -foreground red
    $win tag configure list -spacing1 3p -spacing3 3p
    $win tag configure compact -spacing1 0p
    $win tag configure link -underline 1 -foreground blue

    set_indent $win $var(S_tab)
    $win configure -wrap word
    
    # configure the text insertion point
    $win mark set $var(S_insert) 1.0
    
    # for horizontal rules
    $win tag configure thin -font {times 2 normal} -background black
    $win tag configure hr -relief sunken -borderwidth 2 -wrap none \
	-tabs [winfo width $win]
    bind $win <Configure> {
	%W tag configure hr -tabs %w
	%W tag configure last -spacing3 %h
    }
    
    $win tag bind link <1> [list textHtml::link_hit $win %x %y]
}

# set the indent spacing (in cm) for lists
# TK uses a "weird" tabbing model that causes \t to insert a single
# space if the current line position is past the tab setting

proc textHtml::set_indent {win cm} {
    set tabs [expr $cm / 2.0]
    $win configure -tabs ${tabs}c
    foreach i {1 2 3 4 5 6 7 8 9} {
	set tab [expr $i * $cm]
	$win tag configure indent$i -lmargin1 ${tab}c -lmargin2 ${tab}c \
	    -tabs "[expr $tab + $tabs]c [expr $tab + 2*$tabs]c"
    }
}

# reset the state of window - get ready for the next page
# remove all but the font tags, and remove all form state

proc textHtml::reset_win {win} {
    upvar #0 textHtml::$win var

    regsub -all { +[^L ][^ ]*} " [$win tag names] " {} tags
    catch {$win tag delete $tags}
    eval $win mark unset [$win mark names]
    $win delete 0.0 end
    $win tag configure hr -tabs [winfo width $win]
    
    # configure the text insertion point
    $win mark set $var(S_insert) 1.0
    
    init_state $win
    return textHtml::$win
}

# initialize the window's state array
# Parameters beginning with S_ are NOT reset
#  adjust_size:		global font size adjuster
#  unknown:		character to use for unknown entities
#  tab:			tab stop (in cm)
#  stop:		enabled to stop processing
#  update:		how many tags between update calls
#  tags:		number of tags processed so far
#  symbols:		Symbols to use on un-ordered lists

proc textHtml::init_state {win} {
    upvar #0 textHtml::$win var

    array set tmp [array get var S_*]
    catch {unset var}
    array set var {
	stop 0
	tags 0
	fill 0
	list list
	S_adjust_size 0
	S_tab 1.0
	S_unknown \xb7
	S_update 10
	S_symbols *=+-o\xd7\xb0>:\xb7
	S_insert Insert
    }
    array set var [array get tmp]
}

# alter the parameters of the text state
# this allows an application to over-ride the default settings
# it is called as: set_state -param value -param value ...

array set textHtml::param_map {
    -update S_update
    -tab S_tab
    -unknown S_unknown
    -stop S_stop
    -size S_adjust_size
    -symbols S_symbols
    -insert S_insert
}

proc textHtml::set_state {win args} {
    upvar #0 textHtml::$win var
    variable param_map
    
    set bad 0
    if {[catch {array set params $args}]} {return 0}
	foreach i [array names params] {
	    incr bad [catch {set var($param_map($i)) $params($i)}]
	}
    return [expr $bad == 0]
}

############################################
# manage the display of html

# render gets called for every html tag
#   win:   The name of the text widget to render into
#   tag:   The html tag (in arbitrary case)
#   not:   a "/" or the empty string
#   param: The un-interpreted parameter list
#   text:  The plain text until the next html tag

proc textHtml::render {win tag not param text} {
    upvar #0 textHtml::$win var
    variable tag_map
    variable insert_map
    variable list_elements

    if {$var(stop)} return
    
    set tag [string tolower $tag]
    set text [map_esc $text]
    
    # manage compact rendering of lists
    if {[info exists list_elements($tag)]} {
	set list "list [expr {[extract_param $param compact] ? "compact" : "list"}]"
    } else {
	set list ""
    }
    
    # adjust (push or pop) tag state
    catch {stack $win $not "$tag_map($tag) $list"}
    
    # insert white space (with current font)
    # adding white space can get a bit tricky.  This isn't quite right
    set bad [catch {$win insert $var(S_insert) $insert_map($not$tag) "space $var(font)"}]
    if {!$bad && [lindex $var(fill) end]} {
	set text [string trimleft $text]
    }
    
    # to fill or not to fill
    if {[lindex $var(fill) end]} {
	set text [zap_white $text]
    }
    
    # do any special tag processing
    catch {tag_$not$tag $win $param text} msg
    
     
    # add the text with proper tags
    
    set tags [current_tags $win]
    $win insert $var(S_insert) $text $tags 
    
    # We need to do an update every so often to insure interactive response.
    # This can cause us to re-enter the event loop, and cause recursive
    # invocations of render, so we need to be careful.
    if {!([incr var(tags)] % $var(S_update))} {
	update idletasks
    }
}

# html tags requiring special processing
# Procs of the form tag_<tag> or tag_</tag> get called just before
# the text for this tag is displayed.  These procs are called inside a 
# "catch" so it is OK to fail.
#   win:   The name of the text widget to render into
#   param: The un-interpreted parameter list
#   text:  A pass-by-reference name of the plain text until the next html tag
#          Tag commands may change this to affect what text will be inserted
#          next.

# A pair of pseudo tags are added automatically as the 1st and last html
# tags in the document.  The default is <hmstart> and </hmstart>.
# Append enough blank space at the end of the text widget while
# rendering so goto can place the target near the top of the page,
# then remove the extra space when done rendering.

proc textHtml::tag_hmstart {win param text} {
    upvar #0 textHtml::$win var

    $win mark gravity $var(S_insert) left
    $win mark gravity $var(S_insert) right
}

proc textHtml::tag_/hmstart {win param text} {
    $win delete last.first end
}

# put the document title in the window banner, and remove the title text
# from the document

proc textHtml::tag_title {win param text} {
    upvar 1 $text data
    
    wm title [winfo toplevel $win] $data
    set data ""
}

proc textHtml::tag_hr {win param text} {
    upvar #0 textHtml::$win var
	
    $win insert $var(S_insert) "\n" space "\n" thin "\t" "thin hr" "\n" thin
}

# list element tags

proc textHtml::tag_ol {win param text} {
    upvar #0 textHtml::$win var
    
    set var(count$var(level)) 0
}

proc textHtml::tag_ul {win param text} {
    upvar #0 textHtml::$win var
    
    catch {unset var(count$var(level))}
}

proc textHtml::tag_menu {win param text} {
    upvar #0 textHtml::$win var
	
    set var(menu) ->
    set var(compact) 1
}

proc textHtml::tag_/menu {win param text} {
    upvar #0 textHtml::$win var
	
    catch {unset var(menu)}
    catch {unset var(compact)}
}
	
proc textHtml::tag_dt {win param text} {
    upvar #0 textHtml::$win var
    upvar 1 $text data
	
    set level $var(level)
    incr level -1
    $win insert $var(S_insert) "$data" \
		"hi [lindex $var(list) end] indent$level $var(font)"
    set data {}
}

proc textHtml::tag_li {win param text} {
    upvar #0 textHtml::$win var
	
    set level $var(level)
    incr level -1
    set x [string index $var(S_symbols)+-+-+-+- $level]
    catch {set x [incr var(count$level)]}
    catch {set x $var(menu)}
    $win insert $var(S_insert) \t$x\t "mark [lindex $var(list) end] indent$level $var(font)"
}

# Manage hypertext "anchor" links.  A link can be either a source (href)
# a destination (name) or both.  If its a source, register it via a callback,
# and set its default behavior.  If its a destination, check to see if we need
# to go there now, as a result of a previous goto request.  If so, schedule
# it to happen with the closing </a> tag, so we can highlight the text up to
# the </a>.

proc textHtml::tag_a {win param text} {
    upvar #0 textHtml::$win var
    
    # a source
    
    if {[extract_param $param href]} {
	set var(Tref) [list L:$href]
	stack $win "" "Tlink link"
	link_setup $win $href
    }
    
    # a destination

    if {[extract_param $param name]} {
	set var(Tname) [list N:$name]
	stack $win "" "Tanchor anchor"
	$win mark set N:$name "$var(S_insert) - 1 chars"
	$win mark gravity N:$name left
	if {[info exists var(goto)] && $var(goto) == $name} {
	    unset var(goto)
	    set var(going) $name
	}
    }
}

# The application should call here with the fragment name
# to cause the display to go to this spot.
# If the target exists, go there (and do the callback),
# otherwise schedule the goto to happen when we see the reference.

proc textHtml::goto {win where} {
    upvar #0 textHtml::$win var
	
    if {[regexp N:$where [$win mark names]]} {
	$win see N:$where
	update
	return 1
    } else {
	set var(goto) $where
	return 0
    }
}

proc textHtml::tag_/a {win param text} {
    upvar #0 textHtml::$win var
	
    if {[info exists var(Tref)]} {
	unset var(Tref)
	stack $win / "Tlink link"
    }
    
    # goto this link, then invoke the call-back.

    if {[info exists var(going)]} {
	$win yview N:$var(going)
	unset var(going)
    }

    if {[info exists var(Tname)]} {
	unset var(Tname)
	stack $win / "Tanchor anchor"
    }
}

# Sample hypertext link callback routine - should be replaced by app
# This proc textHtml::is called once for each <A> tag.

#   win:   The name of the text widget to render into
#   href:  The HREF link for this <a> tag.


# We need to escape any %'s in the href tag name so the bind command
# doesn't try to substitute them.

# proc textHtml::link_setup {win href} {
#     #regsub -all {%} $href {%%} href

#     $win tag bind L:$href <Enter> [list $win configure -cursor hand2]
#     $win tag bind L:$href <Leave> [list $win configure -cursor {}]
#     $win tag bind L:$href <1> [list $win configure -cursor {}]
# }

# generic link-hit callback
# This gets called upon button hits on hypertext links
# Applications are expected to supply ther own link_callback routine
#   win:   The name of the text widget to render into
#   x,y:   The cursor position at the "click"

proc textHtml::link_hit {win x y} {
    set tags [$win tag names @$x,$y]
    set link [lindex $tags [lsearch -glob $tags L:*]]
    regsub L: $link {} link
    link_callback $win $link
}

# extract a value from parameter list (this needs a re-do)
# returns "1" if the keyword is found, "0" otherwise
#   param:  A parameter list.  It should alredy have been processed to
#           remove any entity references
#   key:    The parameter name
#   val:    The variable to put the value into (use key as default)

proc textHtml::extract_param {param key {val ""}} {
    
    if {$val == ""} {
	upvar $key result
    } else {
	upvar $val result
    }
    set ws "    \n\r"
    
    # look for name=value combinations.  Either (') or (") are valid delimeters
    if {[regsub -nocase [format {.*%s[%s]*=[%s]*"([^"]*).*} \
	    $key $ws $ws] $param {\1} value] || \
    [regsub -nocase [format {.*%s[%s]*=[%s]*'([^']*).*} $key $ws $ws] \
	 $param {\1} value] || \
    [regsub -nocase [format {.*%s[%s]*=[%s]*([^%s]+).*} $key $ws $ws $ws]\
	 $param {\1} value] } {
set result $value
return 1
}

# now look for valueless names
# I should strip out name=value pairs, so we don't end up with "name"
# inside the "value" part of some other key word - some day
	
set bad \[^a-zA-Z\]+
if {[regexp -nocase  $bad$key$bad -$param-]} {
    return 1
} else {
    return 0
}
}

# These next two routines manage the display state of the page.

# Push or pop tags to/from stack.
# Each orthogonal text property has its own stack, stored as a list.
# The current (most recent) tag is the last item on the list.
# Push is {} for pushing and {/} for popping

proc textHtml::stack {win push list} {
    upvar #0 textHtml::$win var
    
    array set tags $list
    if {$push == ""} {
	foreach tag [array names tags] {
	    lappend var($tag) $tags($tag)
	}
    } else {
	foreach tag [array names tags] {
	    set var($tag) [lreplace $var($tag) end end]
	}
    }
}

# extract set of current text tags
# tags starting with T map directly to text tags, all others are
# handled specially.  There is an application callback, set_font
# to allow the application to do font error handling

proc textHtml::current_tags {win} {
    upvar #0 textHtml::$win var

    set font font
    foreach i {family size weight style} {
	set $i [lindex $var($i) end]
	append font :[set $i]
    }

    catch {incr size $var(S_adjust_size)}
    catch {$win tag configure $font -font "$family $size $weight $style"} 
    set indent [llength $var(indent)]
    incr indent -1
    lappend tags $font indent$indent
    foreach tag [array names var T*] {
	lappend tags [lindex $var($tag) end]
    }
    set var(font) $font
    set var(level) $indent
    return $tags
}

############################################
# Turn HTML into TCL commands
#   html    A string containing an html document
#   cmd		A command to run for each html tag found
#   start	The name of the dummy html start/stop tags

proc textHtml::parse_html {html cmd {start hmstart}} {
    regsub -all \{ $html {\&ob;} html
    regsub -all \} $html {\&cb;} html
    regsub -all {\\} $html {\&bsl;} html
    set w " \t\r\n"
    set exp <(/?)(\[^$w>\]+)\[$w\]*(\[^>\]*)>
    set sub "\}\n$cmd {\\2} {\\1} {\\3} \{"
    regsub -all $exp $html $sub html
    eval "$cmd {$start} {} {} {$html}; $cmd {$start} / {} {}"
}

# Convert multiple white space into a single space

proc textHtml::zap_white {data} {
    regsub -all "\[ \t\r\n\]+" $data " " data
    return $data
}

# find HTML escape characters of the form &xxx;


proc textHtml::map_esc {text} {
    if {![regexp & $text]} {return $text}
    regsub -all {([][$\\])} $text {\\\1} new
    regsub -all {&\#([0-9][0-9]?[0-9]?);?} \
		     $new {[format %c [scan \1 %d tmp;set tmp]]} new
    regsub -all {&([a-zA-Z]+);?} $new {[do_map \1]} new
    return [subst $new]
}

# convert an HTML escape sequence into character

proc textHtml::do_map {text {unknown ?}} {
    variable esc_map
   
    set result $unknown
    catch {set result $esc_map($text)}
    return $result
}


#############################################################
# ISO Latin-1 escape codes

namespace eval textHtml {
    variable esc_map

    # table of escape characters (ISO latin-1 esc's are in a different table)

    array set esc_map {
	lt <   gt >   amp &   quot \"   copy \xa9
	reg \xae   ob \x7b   cb \x7d   nbsp \xa0 bsl \\
    }

    array set esc_map {
	nbsp \xa0 iexcl \xa1 cent \xa2 pound \xa3 curren \xa4
	yen \xa5 brvbar \xa6 sect \xa7 uml \xa8 copy \xa9
	ordf \xaa laquo \xab not \xac shy \xad reg \xae
	hibar \xaf deg \xb0 plusmn \xb1 sup2 \xb2 sup3 \xb3
	acute \xb4 micro \xb5 para \xb6 middot \xb7 cedil \xb8
	sup1 \xb9 ordm \xba raquo \xbb frac14 \xbc frac12 \xbd
	frac34 \xbe iquest \xbf Agrave \xc0 Aacute \xc1 Acirc \xc2
	Atilde \xc3 Auml \xc4 Aring \xc5 AElig \xc6 Ccedil \xc7
	Egrave \xc8 Eacute \xc9 Ecirc \xca Euml \xcb Igrave \xcc
	Iacute \xcd Icirc \xce Iuml \xcf ETH \xd0 Ntilde \xd1
	Ograve \xd2 Oacute \xd3 Ocirc \xd4 Otilde \xd5 Ouml \xd6
	times \xd7 Oslash \xd8 Ugrave \xd9 Uacute \xda Ucirc \xdb
	Uuml \xdc Yacute \xdd THORN \xde szlig \xdf agrave \xe0
	aacute \xe1 acirc \xe2 atilde \xe3 auml \xe4 aring \xe5
	aelig \xe6 ccedil \xe7 egrave \xe8 eacute \xe9 ecirc \xea
	euml \xeb igrave \xec iacute \xed icirc \xee iuml \xef
	eth \xf0 ntilde \xf1 ograve \xf2 oacute \xf3 ocirc \xf4
	otilde \xf5 ouml \xf6 divide \xf7 oslash \xf8 ugrave \xf9
	uacute \xfa ucirc \xfb uuml \xfc yacute \xfd thorn \xfe
	yuml \xff
    }
}
